////////////////////////////////////////////////////////////////////////////////
// Filename: Input_Stream.hh                                                  //
// Purpose.: defines the class 'Input_Stream', a list of Input_Atom's         //
////////////////////////////////////////////////////////////////////////////////
// Author: Markus Keidl                                                       //
// Theme.: 2                                                                  //
// Group.: 5 (Gerald Anleitner, Stefan Seltzsam, Markus Keidl)                //
//                                                                            //
// Date of creation: 20.05.1996        Date of last change: 13.07.1996        //
////////////////////////////////////////////////////////////////////////////////

#ifndef _INPUT_STREAM_HH
#define _INPUT_STREAM_HH

#include <DLList.h>
#include <String.h>
#include <fstream.h>
#include "Input_Atom.hh"

class Input_Stream
{
  friend ostream& operator<<(ostream& s, Input_Stream& str);

public:
  // constructors and destructor
  Input_Stream();		
  Input_Stream(const Input_Stream& rhs);	
  ~Input_Stream();		
  
  // memberfunctions
  void add_Input_Atom(const Input_Atom& atom); // adds an Input_Atom to the list
  int read_from_file(const String& path);      // reads an Input_Stream from a file,
					       // specified by the full file path
  bool is_empty();			       // returns if the stream is empty
  void cut_off(int anz);		       // cuts off the first anz Input_Atom's

  void first();				       // set iterator to first item
  int operator!() const;		       // end of list reached ?
  Input_Atom& operator()();		       // get current item
  void next();				       // set iterator to next item
  Pix get_itr();			       // get current position in list
  void set_itr(Pix i);			       // set new position in list

private:
  DLList<Input_Atom> input_stream;
  Pix itr;			
};

#endif
