////////////////////////////////////////////////////////////////////////////////
// Filename: expand.cc                                                        //
// Purpose : expand a string with valid [..] blocks to full length            //
//           That is used to build up the AUTOMATON, because only |, (, ) ,   //
//           *,+,? are valid operators for AUTOMATON.                         //
//           Also other functions for Automaton are implemented               //
////////////////////////////////////////////////////////////////////////////////
// Author: Stefan Seltzsam                                                    //
// Theme.: 2                                                                  //
// Group.: 5 (Gerald Anleitner, Stefan Seltzsam, Markus Keidl)                //
//                                                                            //
// Date of creation: 26.06.1996        Date of last change: 15.07.1996        //
////////////////////////////////////////////////////////////////////////////////

// is_valid_quoted: returns true, if quote is a valid quote inside (1)
// or outside (0) a [..] block
int Automaton::is_valid_quoted(String quote, int in_out)
{
  String valid_quotes[2] = {"\\[]|*()e?+. ", "\\[]"}; 
                         // outside [..] , inside [..]

  if (valid_quotes[in_out].contains(quote))
    return true;
  else
    return false;
}

int Automaton::check_regexp(String& regexp)
// Check all quotes, if they are valid or not
// Check, if there are only valid [..]-blocks
// Change all \t and \n to '\t' and '\n'
// Change "\ " to " "
// remove all tabs and spaces (no '\ ' or '\n' or '\t')
// Errors: false = Error => Error message is in 'regexp'
{
  String checked_regexp = "";
  int    bracket_counter = 0;

  if (regexp.length() == 0) return true;

  for (unsigned int pos = 0; pos < regexp.length(); )
    {
      switch (regexp[pos])
	{
	// quoted character 
	case '\\':
	  // ERROR: the \ is the last character
	  if (pos == regexp.length()-1) 
	    {
	      regexp = "unquoted \\ in regular expression";
	      return false;
	    }
	  // there could be a \t or \n
	  if (regexp[pos+1] == 't')
	    {
	      checked_regexp += '\t';
	      pos += 2;
	      break;
	    }
	  if (regexp[pos+1] == 'n')
	    {
	      checked_regexp += '\n';
	      pos += 2;
	      break;
	    }
	  // a wrong character might be quoted
	  if (is_valid_quoted(regexp[pos+1],bracket_counter))
	    {
	      if (regexp[pos+1] != ' ')
		checked_regexp += regexp.at(int(pos),2);
	      else
		checked_regexp += ' ';
	      pos += 2;
	      break;
	    }
	  else
	    {
	      regexp = "wrong character quoted in regular expression";
	      return false;
	    }
	  break;
	// Look if character is a [ or ]:
	case '[':
	  bracket_counter++;
	  if (bracket_counter > 1)
	    {
	      regexp = "no unquoted [ allowed inside [..] block in regular expression";
	      return false;
	    }
	  checked_regexp += regexp[pos];
	  pos++;
	  break;
	case ']':
	  bracket_counter--;
	  if (bracket_counter < 0)
	    {
	      regexp = "no unquoted ] allowed outside [..] block in regular expression";
	      return false;
	    }
	  checked_regexp += regexp[pos];
	  pos++;
	  break;
	case '\t':
	case ' ':
	  if (bracket_counter == 1) checked_regexp += " ";
	  pos++;
	  break;
	default:
	  checked_regexp += regexp[pos];
	  pos++;
	}
    }
  // all characters checked, so look if bracket_counter is 0
  if (bracket_counter != 0)
    {
      regexp = "there is a [ without closing ] in regular expression";
      return false;
    }

  // everything is ok
  regexp = checked_regexp;

  return true;
}

// help function: returns position of a valid closing ] (no quoted ]!), starting at start
int Automaton::find_block_end (String& regexp, int start)
{
  int end = start+1;

  while (regexp[end] != ']')
    {
      if (regexp[end] == '\\')
	end += 2;
      else
	end++;
    }

  return end;
}

// returns correct quoted String of characer quote
String Automaton::quote (char quote) 
{
  String do_quote = "\\|*()?+";

  if (do_quote.contains(quote))
    return "\\"+String(quote);
  else
    return String(quote);
}

// expand [..] and [^..]- Blocks to their full length
int Automaton::expand_block (String& block)
{
  String expanded_block = "";  // contains expanded block
  unsigned char from, to, run; // used for a-b parts in blocks
  int startpos;                // 0, if [..] block, 1, if [^..] block
  int initial_value;           // should a character be included (1) or excluded (0) by default?

  if (block[0] != '^')
    {
      startpos = 0;
      initial_value = 0;
    }
  else
    {
      startpos = 1;
      initial_value = 1;
    }
  
  // initialize array
  int characters[256];

  // all is allowed / not allowed
  for (int i = 0; i < 256; i++)
    characters[i] = initial_value;

  // first remove quotes
  for (int pos = 0; pos < int(block.length()); )
    {
      if (block[pos] == '\\')
	{
	  expanded_block += block[pos+1];
	  pos += 2;
	}
      else
	{
	  expanded_block += block[pos];
	  pos++;
	}
    }

  block = expanded_block;
  expanded_block = "";
  
  // now build up regular expression
  for (int pos = startpos; pos < int(block.length()); )
    {
      if (pos < int (block.length() - 2) && block[pos+1] == '-')
	{
	  from = block[pos];
	  to = block[pos+2]; 
	  if (from > to)
	    {
	      block = "illegal order of boundaries in [..] block in regular expression";
	      return false;
	    }
	  
	  for (run = from;; run++)
	    {
	      characters[int(run)] = !initial_value;
	      if (run == to) break;
	    }
	  pos += 3;
	}
      else
	{
	  characters[block[pos]] = !initial_value;
	  pos++;
	}
    }
  
  // now make a String for all allowed characters
  for (int i = 1; i < 256; i++)
    if (characters[i] == 1)
      expanded_block += quote(char(i)) +  "|";
  
  // work is done
  if (expanded_block != "")
    expanded_block = "("+expanded_block.at(0,expanded_block.length()-1)+")";
  block = expanded_block;
  return true;

}


// works on a valid (checked by expand regexp) regular expression
// NOTE: Errors could appear, if you do not check regexp befor using thos function
int Automaton::expand_checked_regexp (String& regexp)
{
  String expanded_regexp = "";
  String block;
  int pos2;

  if (regexp.length() == 0) return true;

  // expand from left to right
  for (int pos = 0; pos < int(regexp.length()); )
    {
      switch (regexp[pos])
	{
	case '\\':
	  switch (regexp[pos+1])
	    {
	    case '[':
	    case ']':
	    case '.':
	      expanded_regexp += regexp[pos+1];
	      break;
	    default:
	      expanded_regexp += regexp.at(pos,2);
	    }
	  pos += 2;
	  break;
	case '[':
	  pos2 = find_block_end(regexp, pos);
	  block = regexp.at(pos+1, pos2-pos-1);
	  if (expand_block (block) == false)
	    {
	      regexp = block;
	      return false;
	    }
	  expanded_regexp += block;
	  pos = pos2+1;
	  break;
	case '.':
	  block = String(char(1))+"-"+String(char(255));
	  expand_block (block);
	  expanded_regexp += block;
	  pos++;
	  break;
	default:
	  expanded_regexp += regexp[pos];
	  pos++;
	}
    }

  regexp = expanded_regexp;
  return true;
}

// expand regexp to full length, and do other things...
int Automaton::expand_regexp (String& regexp)
{
  int error = false;   // false = no error

  // Step 1: check quotes, remove whitespaces, check brackets, replace \t, \n...
  error = check_regexp(regexp);
  if (!error) 
    return false;

  // Step 2: expand checked regexp
  error = expand_checked_regexp(regexp);
#ifdef DEBUG
  cout << "expand_regexp: " << regexp << endl;
#endif 
  if (error) 
    return true;
  else
    return false;

}

// optimize automaton by using other types of transitions
void Automaton::optimize_automaton ()
{
  Transition* flag = NULL;  
  String old_labels;

  DLList <Transition*> tr_list;

  // optimize every state 
  for (Pix i = state_list.first(); i != 0; state_list.next(i))   
    {     
      tr_list.clear();
      // optimize every transition of a state     
      for (state_list(i)->first(); state_list(i)->operator()() != 0; state_list(i)->next())
 	{
 	  flag = NULL; 	  
	  // is actual transition an epsilon transition ?
  	  if (state_list(i)->operator()()->get_label().get_Input_Atom() == "")
 	    {
 	      // if there is already a epsilon transition pointing at the same state, find it.
  	      for (Pix j = tr_list.first(); j != 0; tr_list.next(j)) 		
		{
 		  if (&tr_list(j)->get_targetstate() == &state_list(i)->operator()()->get_targetstate() &&
 		      tr_list(j)->get_label().get_Input_Atom() == "")
 		    flag = tr_list(j);
 		}
 	      if (flag == NULL)
		// there is no equal epsilon transition in tr_list, so add it
 		{ 
		  Transition* new_transition = new Labelled_Transition(*(Labelled_Transition*)state_list(i)->operator()());
		  tr_list.append(new_transition);
 		}
 	    } 	  
	  else 
	    // no epsilon transition 
	    { 	
	      // if there is already a non epsilon transition pointing at the same state, find it.  	      
	      for (Pix j = tr_list.first(); j != 0; tr_list.next(j)) 	
		{
 		  if (&tr_list(j)->get_targetstate() == &state_list(i)->operator()()->get_targetstate() &&
 		      tr_list(j)->get_label().get_Input_Atom() != "")
 		    flag = tr_list(j); 		
		}
 	      if (flag == NULL) 
		// there is no equal non epsilon transition in tr_list, so add it
		{
		  Transition* new_transition = new Labelled_Transition(*(Labelled_Transition*)state_list(i)->operator()());
		  tr_list.append(new_transition);
		}
	      else
		{
		  flag->set_type(OR);
		  old_labels = flag->get_label().get_Input_Atom();
		  if (!old_labels.contains(state_list(i)->operator()()->get_label().get_Input_Atom()))
		    old_labels = old_labels + state_list(i)->operator()()->get_label().get_Input_Atom();
		  if (old_labels.length() == 255)
		    {
		      flag->set_type(POINT);
		      flag->set_label(Input_Atom(""));
		    }
		  else
		    flag->set_label(old_labels);
		}
	    }
	}
    // deleta old transitions and transition_list
    state_list(i)->clear();
    // insert new transitions
    for (Pix k = tr_list.first(); k != 0; tr_list.next(k))
	{
	  state_list(i)->add_transition(tr_list(k));
	}
  }
}


