////////////////////////////////////////////////////////////////////////////////
// Filename: unlabelled_state.cc                                              //
// Purpose : implements the class 'Unlabelled_State'                          //
// class for unlabelled states needed for building up the AUTOMATON           //
////////////////////////////////////////////////////////////////////////////////
// Author: Stefan Seltzsam                                                    //
// Theme.: 2                                                                  //
// Group.: 5 (Gerald Anleitner, Stefan Seltzsam, Markus Keidl)                //
//                                                                            //
// Date of creation: 22.05.1996        Date of last change: 22.05.1996        //
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------

#include <unlabelled_state.hh>

//------------------------------------------------------------------------------
// class Unlabelled_State
//------------------------------------------------------------------------------

// contructor & destructor
Unlabelled_State::Unlabelled_State() 
{
#ifdef DEBUG
  cout << "Unlabelled_State::Unlabelled_State()" << endl;
#endif
  epsilon_loop_detection = this;
  // epsilon-loop-detection is not set to NULL, because Null can be stored, if we are at the
  // end of Input_Stream. So I store this, because the iterator of Input_Stream can not point
  // at me.
};

Unlabelled_State::~Unlabelled_State() 
{  
#ifdef DEBUG
  cout << "Unlabelled_State::~Unlabelled_State()" << endl;
#endif
  for (Pix i = transition_list.first(); i != 0; transition_list.next(i))
    delete transition_list(i);
};

// copy constructor - makes an exact copy of rhs
Unlabelled_State::Unlabelled_State(const Unlabelled_State& rhs)
{
#ifdef DEBUG
  cout << "Unlabelled_State::Unlabelled_State(const Unlabelled_State& rhs)" << endl;
#endif
  transition_list = rhs.transition_list;
  itr = rhs.itr;
  epsilon_loop_detection = rhs.epsilon_loop_detection;
}

// add a transition to transition_list 
void Unlabelled_State::add_transition(Transition* trans)
{
#ifdef DEBUG
  cout << "void Unlabelled_State::add_transition(Transition* trans)" << endl;  
#endif
  transition_list.append(trans);
}

// remove a transition from transition_list 
void Unlabelled_State::remove_transition(Transition* trans)
{
#ifdef DEBUG
  cout << "void Unlabelled_State::remove_transition(Transition* trans)" << endl;  
#endif
  Pix i = transition_list.first();
  while (transition_list(i) != trans && transition_list(i) != NULL)
    transition_list.next(i);

  if (transition_list(i) == trans)
    transition_list.del(i);
}

// set internal iterator to first element of transition_list
void Unlabelled_State::first()
{
#ifdef DEBUG
  cout << "bool Unlabelled_State::first()" << endl;
#endif
  if (!transition_list.empty())
    itr = transition_list.first();
  else
    itr = 0;
}

// set internal iterator to next element of transition_list
void Unlabelled_State::next()
{
#ifdef DEBUG
  cout << "bool Unlabelled_State::next()" << endl; 
#endif
  if (itr != 0)
    transition_list.next(itr);
}

void Unlabelled_State::clear()
{
  for (Pix i = transition_list.first(); i != 0; transition_list.next(i))
    delete transition_list(i);
  transition_list.clear();
  itr = NULL;
}

// returns the actual element of transition_list, which the internal iterator is pointing at
// NULL = End of List
Transition* Unlabelled_State::operator()() const
{
#ifdef DEBUG
  cout << "Transition* Unlabelled_State::operator()() const" << endl;
#endif
  if (itr != 0)
    return transition_list(itr);
  else
    return 0;
}

// forwarding the Protocol_List to all transitions stored in transition_list
inline void Unlabelled_State::send(Protocol& hit)
{
#ifdef DEBUG
  cout << "void Unlabelled_State::send(Protocol_List& hit) const" << endl;  
#endif
  
  // Am I in a epsilon Cycle ?
  if (epsilon_loop_detection == hit.get_Input_Stream().get_itr())
    return;
  
  // No, so store actual itr of Input_Stream
  epsilon_loop_detection = hit.get_Input_Stream().get_itr();

  // Delegate work to all my Transitions
  for (Pix i = transition_list.first(); i != 0; transition_list.next(i))
    transition_list(i)->send(hit);	      
};
